{
=============================================================================
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  |
| AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    |
| IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   |
| ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR  |
| ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL       |
| DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR   |
| SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER   |
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT           |
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY    |
| OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  |
| DAMAGE.                                                                      |
=============================================================================
   Author:  D.J.C. Blijker.
   Date:    23-4-2003
   version: 1.0
   e-mail:  c.blijker@wxs.nl
=============================================================================
  This unit will work in Delphi as well as in Kylix!!!
=============================================================================

 Main target of this unit is to supply applications with an easy-to-use and
 consistent means of finding the exact network-address, port and protocol
 of (self implemented) network services.

 This protocol works in tight cooperation with the ServiceBroker (a windows
 example ServiceBroker application is enclosed.) and the THxServiceProvider
 class.

 Usage:

 Prerequisites:
 -Make sure a ServiceBroker is running somewhere on your local network.
 -Make sure the 'service' you are looking for is using a THxServiceProvider
  class.

 -Instantiate a ServiceSeeker object.
 -Supply the netmask (if it differs from 255.255.255.0)
 -Call SeekService with the applicable parameters and wait for the answer.
  -if the call succeeds the eventhandler "OnAnswer" will be called
   else the eventhandler "OnTimeOut".
==============================================================================
History:

  version: 1.0  Initial release.
==============================================================================
}
unit ServiceSeekerLayer;

interface

uses
{$IFDEF LINUX}
  QExtCtrls,
{$ELSE}
  extctrls,
{$ENDIF}
  SysUtils, HxTcpIpcUdp, HxDataExchange, ServiceBroker_Shared;

type
  TOnAnswer = procedure( sServiceName, sAddress, sPort: string;
                         eProtocol: THxProtocol )of object;

  TOnTimeOut = procedure( sServiceName: string;
                          eProtocol: THxProtocol )of object;

  THxServiceSeeker = class
  private
    procedure HandleTimer(Sender: TObject);
    procedure BroadcastAllOutstandingRequest;
    procedure CheckForTimeOut;
  protected
    udp: THxUdp;
    tmr: TTimer;
    de: THxDataExchange;
    iSec: integer;
    lstServices: THxServiceShortList;
    sNetmask: string;
    FOnAnswer: TOnAnswer;
    FOnTimeOut: TOnTimeOut;
    procedure HandleAnswer( ptr: pchar; iSize: integer;
                            sRemoteIP, sPort: string );
    procedure SetNetMask( sNM: string );
    procedure BroadcastRequest( sNameOfService: string;
                                eProtocol: THxProtocol );
  public
    constructor Create;
    destructor Destroy; override;
    procedure SeekService( sServiceName: string; eProtocol: THxProtocol;
                           iTimeOutInSec: integer = -1 );
    property NetMask: string read sNetMask write SetNetmask;
    property OnAnswer: TOnAnswer read FOnAnswer write FOnAnswer;
    property OnTimeOut: TOnTimeOut read FOnTimeOut write FOnTimeout;
  end;

implementation


{ THxServiceSeeker }

constructor THxServiceSeeker.Create;
begin
  iSec := 0;
// For this service seeker to work we need a valid netmask.!!
  NetMask := '255.255.255.0'; // Best guess..

  udp := THxUdp.Create;
  udp.DestinationPort := SERVICEBROKER_PORT;
  udp.DestinationAddress := GetBroadcastAddress( NetMask );

  udp.OnDataAvailable := HandleAnswer;
  udp.ListeningPort := ANY_FREE_PORT;

  de := THxDataExchange.Create;

  tmr := TTimer.Create( nil );
  tmr.Interval := 1000;
  tmr.Enabled := False;
  tmr.OnTimer := HandleTimer;

  lstServices := THxServiceShortList.Create;
end;

destructor THxServiceSeeker.Destroy;
begin
  udp.Free;
  de.Free;
  lstServices.Free;
  tmr.Free;

  inherited;
end;

procedure THxServiceSeeker.BroadcastAllOutstandingRequest;
var
  i: integer;
begin
  for i := 0 to lstServices.Count -1 do
    BroadcastRequest( lstServices[i]^.sServName, lstServices[i]^.eProtocol );
end;

procedure THxServiceSeeker.CheckForTimeOut;
var
  i: integer;
begin
  i := lstServices.Count -1;
  while i >= 0 do
  begin
    with lstServices[i]^ do
    begin
      if iRemainingTime > 0 then
        Dec( iRemainingTime );
      if iRemainingTime = 0 then
      begin
        if Assigned( FOnTimeOut )then
          FOnTimeOut( sServName, eProtocol );
        lstServices.Delete( i );
      end;
    end;
    Dec( i );
  end;
end;

procedure THxServiceSeeker.HandleTimer( Sender: TObject );
begin
  if lstServices.Count = 0 then
  begin
    tmr.Enabled := False;
    exit;
  end;

  Inc( iSec );
  if iSec mod 5 = 0 then // Every 5 seconds is enough..
    BroadcastAllOutstandingRequest;

  CheckForTimeOut;
end;

procedure THxServiceSeeker.SeekService( sServiceName: string;
                                        eProtocol: THxProtocol;
                                        iTimeOutInSec: integer );
var
  i: integer;
  pServ: PHxServiceShort;
begin
  for i := 0 to lstServices.Count-1 do
  begin
    if( lstServices[i]^.sServName = sServiceName )and
    ( lstServices[i]^.eProtocol = eProtocol )then
    begin
      lstServices[i]^.iRemainingTime := iTimeOutInSec;
      exit; // Already in the list..
    end;
  end;

  new( pServ );
  lstServices.Add( pServ );
  pServ^.sServName := sServiceName;
  pServ^.eProtocol := eProtocol;
  pServ^.iRemainingTime := iTimeOutInSec;

  BroadCastRequest( sServiceName, eProtocol );

  if not( tmr.Enabled )then
    tmr.Enabled := True;
end;

procedure THxServiceSeeker.BroadcastRequest( sNameOfService: string;
                                             eProtocol: THxProtocol );
var
  ptr: pchar;
  iSize: integer;
begin
  de.StartCreatingMessage;
  de.AddString( SAP_HEADER );
  de.AddInteger( SERVICE_REQ );
  de.AddString( sNameOfService );
  de.AddInteger( integer( eProtocol ));
  de.GetCreatedMessage( ptr, iSize );
  udp.SendData( ptr, iSize );
end;

procedure THxServiceSeeker.HandleAnswer( ptr: pchar; iSize: integer;
                                         sRemoteIP, sPort: string );
var
  sHeader, sServiceName, sServiceAddress, sServicePort: string;
  eProtocol: THxProtocol;
  i, iProtocol, iMessNr: integer;
begin
  try
    if iSize > 0 then
      ptr := de.ReadString( ptr, sHeader, iSize )
    else
      exit;

    if( iSize = 0 )or( sHeader <> SAP_HEADER )then
      exit;

    ptr := de.ReadInteger( ptr, iMessNr, iSize );
    if( iSize = 0 )or( iMessNr <> SERVICE_ANS )then
      exit;

    ptr := de.ReadString( ptr, sServiceName, iSize );
    if iSize = 0 then
      exit;

    ptr := de.ReadString( ptr, sServiceAddress, iSize );
    if iSize = 0 then
      exit;

    ptr := de.ReadString( ptr, sServicePort, iSize );
    if iSize = 0 then
      exit;

    de.ReadInteger( ptr, iProtocol, iSize );
    if( iSize <> 0 )or   // Should be ready now..
    ( iProtocol < 0 )or
    ( iProtocol > integer( High( THxProtocol )) )then
      exit;

    eProtocol := THxProtocol( iProtocol );

  // Check whether this was a requested service..
    for i := 0 to lstServices.Count-1 do
    begin
      if( lstServices[i]^.sServName = sServiceName )and
      ( lstServices[i]^.eProtocol = eProtocol )then
      begin
        lstServices.Delete( i );

        if lstServices.Count = 0 then
          tmr.Enabled := False;

        if Assigned( FOnAnswer )then
          FOnAnswer( sServiceName, sServiceAddress, sServicePort, eProtocol );

        break;
      end;
    end;
  except;
  // ignore errors..
  end;
end;

procedure THxServiceSeeker.SetNetMask( sNM: string );
var
  s: string;
begin
  s := ValidateIpAddress( sNM );
  if s <> '' then
    sNetMask := s
  else
    raise Exception.Create( sNM+' is not a valid netmask!');
end;


end.
